from __future__ import annotations

import asyncio
import contextlib
import logging
import typing as t

import aiohttp
import anyio
import httpx
from pyparsing import cast
from starlette.requests import Request

from _bentoml_sdk import Service
from bentoml import get_current_service
from bentoml import server_context
from bentoml._internal.utils import expand_envs
from bentoml.exceptions import BentoMLConfigException

if t.TYPE_CHECKING:
    from anyio.abc._subprocesses import Process
    from starlette.applications import Starlette

logger = logging.getLogger("bentoml.server")


async def _check_health(client: aiohttp.ClientSession, health_endpoint: str) -> bool:
    try:
        response = await client.get(health_endpoint, timeout=aiohttp.ClientTimeout(5))
        if response.status == 404:
            raise BentoMLConfigException(
                f"Health endpoint {health_endpoint} not found (404). Please make sure the health "
                "endpoint is correctly configured in the service config."
            )
        return response.ok
    except aiohttp.ClientError:
        return False


def create_proxy_app(service: Service[t.Any]) -> Starlette:
    """A reverse-proxy that forwards all requests to the HTTP server started
    by the custom command.
    """
    import fastapi
    from fastapi.responses import StreamingResponse

    health_endpoint = service.config.get("endpoints", {}).get("livez", "/health")

    @contextlib.asynccontextmanager
    async def lifespan(
        app: fastapi.FastAPI,
    ) -> t.AsyncGenerator[dict[str, t.Any], None]:
        server_instance = get_current_service()
        assert server_instance is not None, "Current service is not initialized"
        # Only start process for the first worker or when replicate_process is True
        should_start_process = (
            service.config.get("replicate_process", False)
            or server_context.worker_index == 1
        )
        async with contextlib.AsyncExitStack() as stack:
            proxy_port = service.config.get("http", {}).get("proxy_port", 8000)
            proxy_url = f"http://localhost:{proxy_port}"
            proc: Process | None = None
            if (
                instance_client := getattr(server_instance, "client", None)
            ) is not None and isinstance(instance_client, aiohttp.ClientSession):
                # TODO: support aiohttp client
                client = instance_client
            else:
                client = await stack.enter_async_context(
                    aiohttp.ClientSession(
                        base_url=proxy_url,
                        timeout=aiohttp.ClientTimeout(),
                        connector=aiohttp.TCPConnector(limit=512),
                    )
                )
            if should_start_process:
                if cmd_getter := getattr(server_instance, "__command__", None):
                    if not callable(cmd_getter):
                        raise TypeError(
                            f"__command__ must be a callable that returns a list of strings, got {type(cmd_getter)}"
                        )
                    cmd = cast("list[str]", cmd_getter())
                else:
                    cmd = service.cmd
                assert cmd is not None, "must have a command"
                cmd = [expand_envs(c) for c in cmd]
                logger.info("Running service with command: %s", " ".join(cmd))
                proc = await anyio.open_process(cmd, stdout=None, stderr=None)
            while proc is None or proc.returncode is None:
                if await _check_health(client, health_endpoint):
                    break
                await anyio.sleep(0.5)
            else:
                raise RuntimeError(
                    "Service process exited before becoming healthy, see the error above"
                )

            app.state.client = client
            try:
                state = {
                    "proc": proc,
                    "client": await stack.enter_async_context(
                        httpx.AsyncClient(base_url=proxy_url, timeout=None)
                    ),  # For backward compatibility
                }
                service.context.state.update(state)
                yield state
            finally:
                if proc is not None:
                    proc.terminate()
                    try:
                        await asyncio.wait_for(proc.wait(), timeout=30.0)
                    except asyncio.TimeoutError:
                        proc.kill()
                        await proc.wait()

    assert service.has_custom_command(), "Service does not have custom command"
    app = fastapi.FastAPI(lifespan=lifespan)

    # TODO: support websocket endpoints
    @app.api_route(
        "/{path:path}",
        methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "HEAD"],
    )
    async def reverse_proxy(request: Request, path: str):
        from contextlib import AsyncExitStack

        import yarl
        from starlette.background import BackgroundTask

        url = yarl.URL.build(path=f"/{path}", query=request.url.query or None)
        client = t.cast(aiohttp.ClientSession, app.state.client)
        headers = dict(request.headers)
        headers.pop("host", None)
        stack = AsyncExitStack()
        try:
            resp = await stack.enter_async_context(
                client.request(
                    url=url,
                    method=request.method,
                    headers=headers,
                    data=request.stream(),
                )
            )
        except aiohttp.ClientError as e:
            logger.error("Client error occurred while proxying request: %s", str(e))
            await stack.aclose()
            return fastapi.Response(
                content="Upstream service unavailable.", status_code=503
            )
        except Exception:
            logger.exception("Error occurred while proxying request")
            await stack.aclose()
            return fastapi.Response(status_code=500)

        return StreamingResponse(
            resp.content.iter_any(),
            status_code=resp.status,
            headers=resp.headers,
            background=BackgroundTask(stack.aclose),
        )

    return app



from __future__ import annotations

import typing as t

import click
import rich

import bentoml.api_token
from bentoml.exceptions import BentoMLException
from bentoml_cli.utils import BentoMLCommandGroup


@click.group(name="api-token", cls=BentoMLCommandGroup)
def api_token_command():
    """API Token management commands."""


@api_token_command.command(name="list")
@click.option(
    "--search", type=click.STRING, default=None, help="Search for tokens by name."
)
@click.option(
    "-o",
    "--output",
    help="Display the output of this command.",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
def list_api_tokens(
    search: str | None,
    output: str,
) -> None:
    """List all API tokens on BentoCloud."""
    import json as json_mod

    import yaml
    from rich.syntax import Syntax
    from rich.table import Table

    try:
        tokens = bentoml.api_token.list(search=search)
    except BentoMLException as e:
        _raise_api_token_error(e, "list")

    if output == "table":
        table = Table(box=None, expand=True)
        table.add_column("Name", overflow="fold")
        table.add_column("UID", overflow="fold")
        table.add_column("Created_At", overflow="fold")
        table.add_column("Expired_At", overflow="fold")
        table.add_column("Last_Used_At", overflow="fold")
        table.add_column("Scopes", overflow="fold")

        for token in tokens:
            expired_at = (
                token.expired_at.strftime("%Y-%m-%d %H:%M:%S")
                if token.expired_at
                else "Never"
            )
            last_used_at = (
                token.last_used_at.strftime("%Y-%m-%d %H:%M:%S")
                if token.last_used_at
                else "Never"
            )
            table.add_row(
                token.name,
                token.uid,
                token.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                expired_at,
                last_used_at,
                ", ".join(token.scopes) if token.scopes else "-",
            )
        rich.print(table)
    elif output == "json":
        res: t.List[dict[str, t.Any]] = [t.to_dict() for t in tokens]
        info = json_mod.dumps(res, indent=2, default=str)
        rich.print(info)
    elif output == "yaml":
        res: t.List[dict[str, t.Any]] = [t.to_dict() for t in tokens]
        info = yaml.dump(res, indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


def _parse_expiration(expires_str: str | None) -> "t.Any":
    """Parse expiration string into datetime."""
    from datetime import datetime
    from datetime import timedelta

    if not expires_str:
        return None

    expires_str = expires_str.strip()

    # Try parsing duration format (e.g., 30d, 1w, 24h)
    if expires_str[-1].lower() in ("d", "w", "h"):
        try:
            value = int(expires_str[:-1])
            unit = expires_str[-1].lower()
            if unit == "d":
                return datetime.now() + timedelta(days=value)
            elif unit == "w":
                return datetime.now() + timedelta(weeks=value)
            elif unit == "h":
                return datetime.now() + timedelta(hours=value)
        except ValueError:
            pass

    # Try parsing ISO date format
    for fmt in ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"):
        try:
            return datetime.strptime(expires_str, fmt)
        except ValueError:
            continue

    raise click.BadParameter(
        f"Invalid expiration format: {expires_str}. "
        "Use duration (e.g., '30d', '1w', '24h') or date (e.g., '2024-12-31')."
    )


@api_token_command.command(name="create")
@click.argument(
    "name",
    nargs=1,
    type=click.STRING,
    required=True,
)
@click.option(
    "-d",
    "--description",
    type=click.STRING,
    help="Description for the API token.",
)
@click.option(
    "--scope",
    "-s",
    type=click.STRING,
    multiple=True,
    help="Scopes for the token (can be specified multiple times). "
    "Available scopes: api, read_organization, write_organization, read_cluster, write_cluster.",
)
@click.option(
    "--expires",
    type=click.STRING,
    help="Expiration time (e.g., '30d' for 30 days, '1w' for 1 week, or ISO date '2024-12-31').",
)
@click.option(
    "-o",
    "--output",
    help="Display the output of this command.",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
def create_api_token(
    name: str,
    description: str | None,
    scope: tuple[str, ...],
    expires: str | None,
    output: str,
) -> None:
    """Create a new API token on BentoCloud.

    \b
    Available scopes:
      - api: General API access
      - read_organization: Read organization data
      - write_organization: Write organization data
      - read_cluster: Read cluster data
      - write_cluster: Write cluster data

    \b
    Examples:
      bentoml api-token create my-token --scope api --scope read_cluster
      bentoml api-token create my-token -s api -s write_organization --expires 30d
    """
    import json as json_mod

    import yaml
    from rich.panel import Panel
    from rich.syntax import Syntax

    expired_at = _parse_expiration(expires)
    scopes = list(scope) if scope else None

    try:
        token = bentoml.api_token.create(
            name=name,
            description=description,
            scopes=scopes,
            expired_at=expired_at,
        )
    except BentoMLException as e:
        _raise_api_token_error(e, "create")

    # Display the token value prominently since it's only shown once
    if token.token:
        rich.print(
            Panel(
                f"[bold green]{token.token}[/bold green]",
                title="[bold yellow]API Token (save this - it won't be shown again!)[/bold yellow]",
                border_style="yellow",
            )
        )

    if output == "table":
        rich.print(f"\nToken [green]{token.name}[/] created successfully")
        rich.print(f"  UID: {token.uid}")
        if token.description:
            rich.print(f"  Description: {token.description}")
        if token.expired_at:
            rich.print(f"  Expires: {token.expired_at.strftime('%Y-%m-%d %H:%M:%S')}")
        else:
            rich.print("  Expires: Never")
    elif output == "json":
        info = json_mod.dumps(token.to_dict(), indent=2, default=str)
        rich.print(info)
    elif output == "yaml":
        info = yaml.dump(token.to_dict(), indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


@api_token_command.command(name="get")
@click.argument(
    "token_uid",
    nargs=1,
    type=click.STRING,
    required=True,
)
@click.option(
    "-o",
    "--output",
    help="Display the output of this command.",
    type=click.Choice(["json", "yaml", "table"]),
    default="table",
)
def get_api_token(token_uid: str, output: str) -> None:
    """Get an API token by UID from BentoCloud."""
    import json as json_mod

    import yaml
    from rich.syntax import Syntax
    from rich.table import Table

    try:
        token = bentoml.api_token.get(token_uid=token_uid)
    except BentoMLException as e:
        _raise_api_token_error(e, "get")

    if token is None:
        raise BentoMLException(f"API token with UID '{token_uid}' not found")

    if output == "table":
        table = Table(box=None, expand=True)
        table.add_column("Field", style="bold")
        table.add_column("Value", overflow="fold")

        table.add_row("Name", token.name)
        table.add_row("UID", token.uid)
        table.add_row("Description", token.description or "-")
        table.add_row("Created At", token.created_at.strftime("%Y-%m-%d %H:%M:%S"))
        table.add_row(
            "Expired At",
            token.expired_at.strftime("%Y-%m-%d %H:%M:%S")
            if token.expired_at
            else "Never",
        )
        table.add_row(
            "Last Used At",
            token.last_used_at.strftime("%Y-%m-%d %H:%M:%S")
            if token.last_used_at
            else "Never",
        )
        table.add_row("Is Expired", str(token.is_expired))
        table.add_row("Scopes", ", ".join(token.scopes) if token.scopes else "-")
        table.add_row("Created By", token.created_by or token.user.name)
        rich.print(table)
    elif output == "json":
        info = json_mod.dumps(token.to_dict(), indent=2, default=str)
        rich.print(info)
    elif output == "yaml":
        info = yaml.dump(token.to_dict(), indent=2, sort_keys=False)
        rich.print(Syntax(info, "yaml", background_color="default"))


@api_token_command.command(name="delete")
@click.argument(
    "token_uid",
    nargs=1,
    type=click.STRING,
    required=True,
)
def delete_api_token(token_uid: str) -> None:
    """Delete an API token on BentoCloud."""
    try:
        bentoml.api_token.delete(token_uid=token_uid)
        rich.print(f"API token [green]{token_uid}[/] deleted successfully")
    except BentoMLException as e:
        _raise_api_token_error(e, "delete")


def _raise_api_token_error(err: "BentoMLException", action: str) -> "t.NoReturn":
    from http import HTTPStatus

    if err.error_code == HTTPStatus.UNAUTHORIZED:
        raise BentoMLException(
            f"{err}\n* BentoCloud API token is required for authorization. Run `bentoml cloud login` command to login"
        ) from None
    elif err.error_code == HTTPStatus.NOT_FOUND:
        raise BentoMLException("API token not found") from None
    raise BentoMLException(f"Failed to {action} API token: {err}")



from __future__ import annotations

import typing as t
from datetime import datetime

import attr
import yaml

from ..utils.cattr import bentoml_cattr
from .schemas.schemasv1 import ApiTokenSchema
from .schemas.schemasv1 import CreateApiTokenSchema

if t.TYPE_CHECKING:
    from .client import RestApiClient


@attr.define(kw_only=True)
class ApiToken(ApiTokenSchema):
    created_by: str = attr.field(default="")

    def to_dict(self) -> dict[str, t.Any]:
        return bentoml_cattr.unstructure(self)

    def to_yaml(self) -> str:
        dt = self.to_dict()
        return yaml.dump(dt, sort_keys=False)

    @classmethod
    def from_schema(cls, schema: ApiTokenSchema) -> ApiToken:
        return cls(
            name=schema.name,
            uid=schema.uid,
            resource_type=schema.resource_type,
            labels=schema.labels,
            created_at=schema.created_at,
            updated_at=schema.updated_at,
            deleted_at=schema.deleted_at,
            description=schema.description,
            scopes=schema.scopes,
            user=schema.user,
            organization=schema.organization,
            expired_at=schema.expired_at,
            last_used_at=schema.last_used_at,
            is_expired=schema.is_expired,
            is_api_token=schema.is_api_token,
            is_organization_token=schema.is_organization_token,
            is_global_access=schema.is_global_access,
            token=schema.token,
            created_by=schema.user.name,
        )


@attr.define
class ApiTokenAPI:
    _client: RestApiClient

    def list(self, search: str | None = None) -> t.List[ApiToken]:
        """
        List all API tokens.

        Args:
            search (str | None): Optional search string to filter tokens.

        Returns:
            List[ApiToken]: A list of ApiToken objects.
        """
        tokens = self._client.v1.list_api_tokens(search=search)
        return [ApiToken.from_schema(token) for token in tokens.items]

    def create(
        self,
        name: str,
        description: str | None = None,
        scopes: t.List[str] | None = None,
        expired_at: datetime | None = None,
    ) -> ApiToken:
        """
        Create a new API token.

        Args:
            name (str): Name of the token.
            description (str | None): Description of the token.
            scopes (List[str] | None): List of scopes for the token.
            expired_at (datetime | None): Expiration datetime for the token.

        Returns:
            ApiToken: The created API token (includes the token value).
        """
        schema = CreateApiTokenSchema(
            name=name,
            description=description,
            scopes=scopes,
            expired_at=expired_at,
        )
        token = self._client.v1.create_api_token(schema)
        return ApiToken.from_schema(token)

    def get(self, token_uid: str) -> ApiToken | None:
        """
        Get an API token by UID.

        Args:
            token_uid (str): The UID of the token to get.

        Returns:
            ApiToken | None: The API token if found, None otherwise.
        """
        token = self._client.v1.get_api_token(token_uid)
        if token is None:
            return None
        return ApiToken.from_schema(token)

    def delete(self, token_uid: str) -> None:
        """
        Delete an API token.

        Args:
            token_uid (str): The UID of the token to delete.
        """
        self._client.v1.delete_api_token(token_uid)



from __future__ import annotations

import contextlib
import logging
import os
import secrets
import threading
import typing as t
from datetime import datetime
from datetime import timezone
from functools import lru_cache
from functools import wraps
from reprlib import recursive_repr as _recursive_repr
from typing import TYPE_CHECKING

import attr
import httpx
from simple_di import Provide
from simple_di import inject

from ...configuration import get_debug_mode
from ...configuration.containers import BentoMLContainer
from .schemas import CommonProperties
from .schemas import EventMeta
from .schemas import ServeInitEvent
from .schemas import ServeUpdateEvent
from .schemas import TrackingPayload

if TYPE_CHECKING:
    from prometheus_client.samples import Sample
    from typing_extensions import Self

    from _bentoml_sdk import Service as NewService
    from bentoml.legacy import Service

    from ...server.metrics.prometheus import PrometheusClient

    P = t.ParamSpec("P")
    T = t.TypeVar("T")
    F = t.Callable[P, t.Any]
    AsyncFunc = t.Callable[P, t.Coroutine[t.Any, t.Any, t.Any]]

logger = logging.getLogger(__name__)

BENTOML_DO_NOT_TRACK = "BENTOML_DO_NOT_TRACK"
BENTOML_SERVE_FROM_SERVER_API = "__BENTOML_SERVE_FROM_SERVER_API"
USAGE_TRACKING_URL = "https://t.bentoml.com"
SERVE_USAGE_TRACKING_INTERVAL_SECONDS = int(12 * 60 * 60)  # every 12 hours
USAGE_REQUEST_TIMEOUT_SECONDS = 1


@lru_cache(maxsize=None)
def _bentoml_serve_from_server_api() -> bool:
    return os.environ.get(BENTOML_SERVE_FROM_SERVER_API, "False").lower() == "true"


@lru_cache(maxsize=1)
def do_not_track() -> bool:  # pragma: no cover
    # Returns True if and only if the environment variable is defined and has value True.
    # The function is cached for better performance.
    return os.environ.get(BENTOML_DO_NOT_TRACK, "False").lower() == "true"


@lru_cache(maxsize=1)
def _usage_event_debugging() -> bool:
    # For BentoML developers only - debug and print event payload if turned on
    return os.environ.get("__BENTOML_DEBUG_USAGE", "False").lower() == "true"


def silent(func: t.Callable[P, T]) -> t.Callable[P, T]:  # pragma: no cover
    # Silent errors when tracking
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
        try:
            return func(*args, **kwargs)
        except Exception as err:  # pylint: disable=broad-except
            if _usage_event_debugging():
                if get_debug_mode():
                    logger.error(
                        "Tracking Error: %s", err, stack_info=True, stacklevel=3
                    )
                else:
                    logger.info("Tracking Error: %s", err)
            else:
                logger.debug("Tracking Error: %s", err)

    return wrapper


@attr.define
class ServeInfo:
    serve_id: str
    serve_started_timestamp: datetime


def get_serve_info() -> ServeInfo:  # pragma: no cover
    # Returns a safe token for serve as well as timestamp of creating this token
    return ServeInfo(
        serve_id=secrets.token_urlsafe(32),
        serve_started_timestamp=datetime.now(timezone.utc),
    )


@inject
def get_payload(
    event_properties: EventMeta,
    session_id: str = Provide[BentoMLContainer.session_id],
) -> t.Dict[str, t.Any]:
    return TrackingPayload(
        session_id=session_id,
        common_properties=CommonProperties(),
        event_properties=event_properties,
        event_type=event_properties.event_name,
    ).to_dict()


@silent
def track(event_properties: EventMeta):
    if do_not_track():
        return
    payload = get_payload(event_properties=event_properties)

    if _usage_event_debugging():
        # For internal debugging purpose
        logger.info("Tracking Payload: %s", payload)
        return

    httpx.post(USAGE_TRACKING_URL, json=payload, timeout=USAGE_REQUEST_TIMEOUT_SECONDS)


@inject
def _track_serve_init(
    svc: Service | NewService[t.Any],
    production: bool,
    serve_kind: str,
    from_server_api: bool,
    serve_info: ServeInfo = Provide[BentoMLContainer.serve_info],
):
    from bentoml.legacy import Service

    is_legacy = isinstance(svc, Service)

    if svc.bento is not None:
        bento = svc.bento
        apis = getattr(bento.info, "apis", [])
        event_properties = ServeInitEvent(
            serve_id=serve_info.serve_id,
            serve_from_bento=True,
            serve_from_server_api=from_server_api,
            production=production,
            serve_kind=serve_kind,
            bento_creation_timestamp=bento.info.creation_time,
            num_of_models=len(bento.info.all_models),
            num_of_runners=len(svc.runners) if is_legacy else len(svc.dependencies),
            num_of_apis=len(apis),
            model_types=[m.module for m in bento.info.all_models],
            runnable_types=[
                r.runnable_type for r in getattr(bento.info, "runners", [])
            ],
            api_input_types=[api.input_type for api in apis],
            api_output_types=[api.output_type for api in apis],
        )
    else:
        if is_legacy:
            num_models = len(
                set(
                    svc.models
                    + [model for runner in svc.runners for model in runner.models]
                )
            )
        else:

            def _get_models(svc: NewService[t.Any], seen: set[str]) -> set[int]:
                if svc.name in seen:
                    return set()
                seen.add(svc.name)
                # Fix: Ensure models are hashable by using id()
                models = {id(model) for model in svc.models}
                for dep in svc.dependencies.values():
                    if dep.on is not None:
                        models.update(_get_models(dep.on, seen))
                return models

            num_models = len(_get_models(svc, set()))

        event_properties = ServeInitEvent(
            serve_id=serve_info.serve_id,
            serve_from_bento=False,
            serve_from_server_api=from_server_api,
            production=production,
            serve_kind=serve_kind,
            bento_creation_timestamp=None,
            num_of_models=num_models,
            num_of_runners=len(svc.runners) if is_legacy else len(svc.dependencies),
            num_of_apis=len(svc.apis.keys()),
            runnable_types=(
                [r.runnable_class.__name__ for r in svc.runners]
                if is_legacy
                else [d.on.name for d in svc.dependencies.values() if d.on is not None]
            ),
            api_input_types=(
                [api.input.__class__.__name__ for api in svc.apis.values()]
                if is_legacy
                else []
            ),
            api_output_types=(
                [api.output.__class__.__name__ for api in svc.apis.values()]
                if is_legacy
                else []
            ),
        )

    track(event_properties)


EXCLUDE_PATHS = {"/docs.json", "/livez", "/healthz", "/readyz"}


class compose:
    """
    Function composition: compose(f, g)(...) is equivalent to f(g(...)).
    Refer to https://github.com/mentalisttraceur/python-compose for original implementation.

    Args:
        *functions: Functions (or other callables) to compose.

    Raises:
        TypeError: If no arguments are given, or any argument is not callable.
    """

    def __init__(self: Self, *functions: F[t.Any]):
        if not functions:
            raise TypeError(f"{self!r} needs at least one argument.")
        _functions: list[F[t.Any]] = []
        for function in reversed(functions):
            if not callable(function):
                raise TypeError(f"{self!r} arguments must be callable.")
            if isinstance(function, compose):
                _functions.extend(function.functions)
            else:
                _functions.append(function)
        self.__wrapped__ = _functions[0]
        self._wrappers = tuple(_functions[1:])

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """Call the composed function."""
        result = self.__wrapped__(*args, **kwargs)
        for function in self._wrappers:
            result = function(result)
        return result

    def __get__(self, obj: t.Any, typ_: type | None = None):
        """Get the composed function as a bound method."""
        wrapped = self.__wrapped__
        try:
            bind = type(wrapped).__get__
        except AttributeError:
            return self
        bound_wrapped = bind(wrapped, obj, typ_)
        if bound_wrapped is wrapped:
            return self
        bound_self = type(self)(bound_wrapped)
        bound_self._wrappers = self._wrappers
        return bound_self

    @_recursive_repr("<...>")
    def __repr__(self):
        return f"{self!r}({','.join(map(repr, reversed(self.functions)))})"

    @property
    def functions(self):
        """Read-only tuple of the composed callables, in order of execution."""
        return (self.__wrapped__,) + tuple(self._wrappers)


def filter_metrics(
    samples: list[Sample], *filters: t.Callable[[list[Sample]], list[Sample]]
):
    return [
        {**sample.labels, "value": sample.value}
        for sample in compose(*filters)(samples)
    ]


def get_metrics_report(
    metrics_client: PrometheusClient,
    serve_kind: str,
) -> list[dict[str, str | float]]:
    """
    Get Prometheus metrics reports from the metrics client. This will be used to determine tracking events.
    If the return metrics are legacy metrics, the metrics will have prefix BENTOML_, otherwise they will have prefix bentoml_

    Args:
        metrics_client: Instance of bentoml._internal.server.metrics.prometheus.PrometheusClient
        grpc: Whether the metrics are for gRPC server.

    Returns:
        A tuple of a list of metrics and an optional boolean to determine whether the return metrics are legacy metrics.
    """
    for metric in metrics_client.text_string_to_metric_families():
        metric_type = t.cast("str", metric.type)  # type: ignore (we need to cast due to no prometheus types)
        metric_name = t.cast("str", metric.name)  # type: ignore (we need to cast due to no prometheus types)
        metric_samples = t.cast("list[Sample]", metric.samples)  # type: ignore (we need to cast due to no prometheus types)
        if metric_type != "counter":
            continue
        # We only care about the counter metrics.
        assert metric_type == "counter"
        if serve_kind == "grpc":
            _filters: list[t.Callable[[list[Sample]], list[Sample]]] = [
                lambda samples: [s for s in samples if "api_name" in s.labels]
            ]
        elif serve_kind == "http":
            _filters = [
                lambda samples: [
                    s
                    for s in samples
                    if not s.labels["endpoint"].startswith("/static_content/")
                ],
                lambda samples: [
                    s for s in samples if s.labels["endpoint"] not in EXCLUDE_PATHS
                ],
                lambda samples: [s for s in samples if "endpoint" in s.labels],
            ]
        else:
            raise NotImplementedError("Unknown serve kind %s" % serve_kind)
        # If metrics prefix is BENTOML_, this is legacy metrics
        if metric_name.endswith("_request") and (
            metric_name.startswith("bentoml_") or metric_name.startswith("BENTOML_")
        ):
            return filter_metrics(metric_samples, *_filters)

    return []


@inject
@contextlib.contextmanager
def track_serve(
    svc: Service | NewService[t.Any],
    *,
    production: bool = False,
    from_server_api: bool | None = None,
    serve_kind: str = "http",
    component: str = "standalone",
    metrics_client: PrometheusClient = Provide[BentoMLContainer.metrics_client],
    serve_info: ServeInfo = Provide[BentoMLContainer.serve_info],
) -> t.Generator[None, None, None]:
    if do_not_track():
        yield
        return

    if from_server_api is None:
        from_server_api = _bentoml_serve_from_server_api()

    _track_serve_init(
        svc=svc,
        production=production,
        serve_kind=serve_kind,
        from_server_api=from_server_api,
    )

    if _usage_event_debugging():
        tracking_interval = 5
    else:
        tracking_interval = SERVE_USAGE_TRACKING_INTERVAL_SECONDS  # pragma: no cover

    stop_event = threading.Event()

    @silent
    def loop() -> t.NoReturn:  # type: ignore
        last_tracked_timestamp: datetime = serve_info.serve_started_timestamp
        while not stop_event.wait(tracking_interval):  # pragma: no cover
            now = datetime.now(timezone.utc)
            event_properties = ServeUpdateEvent(
                serve_id=serve_info.serve_id,
                production=production,
                # Note that we are currently only have two tracking jobs: http and grpc
                serve_kind=serve_kind,
                # Current accept components are "standalone", "api_server" and "runner"
                component=component,
                # check if serve is running from server API or just normal CLI
                serve_from_server_api=from_server_api,
                triggered_at=now,
                duration_in_seconds=int((now - last_tracked_timestamp).total_seconds()),
                metrics=get_metrics_report(metrics_client, serve_kind=serve_kind),
            )
            last_tracked_timestamp = now
            track(event_properties)

    tracking_thread = threading.Thread(target=loop, daemon=True)
    try:
        tracking_thread.start()
        yield
    finally:
        stop_event.set()
        tracking_thread.join()



from __future__ import annotations

import asyncio
import functools
import inspect
import io
import json
import logging
import os
import pathlib
import tempfile
import time
import typing as t
from functools import cached_property
from http import HTTPStatus
from urllib.parse import urljoin
from urllib.parse import urlparse

import aiohttp
import anyio.from_thread
import attr
from multidict import CIMultiDict

from _bentoml_sdk.service.config import ServiceConfig
from bentoml import __version__
from bentoml._internal.io_descriptors.base import IODescriptor
from bentoml._internal.utils.uri import uri_to_path
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import NotFound
from bentoml.exceptions import ServiceUnavailable

from ..serde import Payload
from .base import AbstractClient
from .base import ClientEndpoint
from .base import ClientFileManager
from .task import AsyncTask
from .task import ResultStatus
from .task import Task

T = t.TypeVar("T")
if t.TYPE_CHECKING:
    from PIL import Image

    P = t.ParamSpec("P")

    from _bentoml_sdk import Service
    from bentoml._internal.external_typing import ASGIApp

    from ..serde import Serde


logger = logging.getLogger("bentoml.io")


async def map_exception(resp: aiohttp.ClientResponse) -> BentoMLException:
    status = HTTPStatus(resp.status)
    exc = BentoMLException.error_mapping.get(status, BentoMLException)
    return exc(await resp.text(), error_code=status)


class SessionManager:
    def __init__(
        self,
        url: str,
        timeout: float,
        headers: dict[str, str],
        app: ASGIApp | None = None,
        max_age: float = 300.0,
        max_requests: int = 100,
    ) -> None:
        self.max_age = max_age
        self.max_requests = max_requests

        self._session: aiohttp.ClientSession | None = None
        self._created_at = 0.0
        self._request_count = 0

        parsed = urlparse(url)
        connector: aiohttp.BaseConnector | None = None
        if app is not None:
            from aiohttp_asgi_connector import ASGIApplicationConnector

            connector = ASGIApplicationConnector(app)  # type: ignore[arg-type]
            url = "http://127.0.0.1:3000"
        elif parsed.scheme == "file":
            from aiohttp import UnixConnector

            connector = UnixConnector(path=uri_to_path(url))
            url = "http://127.0.0.1:3000"
        elif parsed.scheme == "tcp":
            url = f"http://{parsed.netloc}"
        self._make_client = lambda: aiohttp.ClientSession(
            base_url=url,
            headers=headers,
            timeout=aiohttp.ClientTimeout(total=timeout),
            connector=connector,
        )

    def _should_refresh(self) -> bool:
        if (time.monotonic() - self._created_at) > self.max_age:
            return True
        if self._request_count > self.max_requests:
            return True
        return False

    async def get_session(self) -> aiohttp.ClientSession:
        if self._session is None or self._should_refresh():
            if self._session is not None:
                await self._session.close()
            self._session = self._make_client()
            self._created_at = time.monotonic()
            self._request_count = 0
        self._request_count += 1
        return self._session

    async def close(self) -> None:
        if self._session is not None:
            await self._session.close()
            self._session = None


@attr.define(slots=False)
class AsyncClient(AbstractClient):
    url: str
    endpoints: dict[str, ClientEndpoint] = attr.field(factory=dict)
    media_type: str = "application/json"
    timeout: float = 30
    default_headers: dict[str, str] = attr.field(factory=dict)
    app: ASGIApp | None = None
    server_ready_timeout: float | None = None
    service: Service[t.Any] | None = None

    _file_manager: ClientFileManager = attr.field(init=False, factory=ClientFileManager)
    _temp_dir: tempfile.TemporaryDirectory[str] = attr.field(init=False)
    _setup_done: bool = attr.field(init=False, default=False)

    def __init__(
        self,
        url: str,
        *,
        media_type: str = "application/json",
        service: Service[t.Any] | None = None,
        server_ready_timeout: float | None = None,
        token: str | None = None,
        timeout: float = 30,
        max_age: float = 300.0,
        max_requests: int = 100,
        app: ASGIApp | None = None,
    ) -> None:
        """Create a client instance from a URL.

        Args:
            url: The URL of the BentoML service.
            media_type: The media type to use for serialization. Defaults to
                "application/json".

        .. note::

            The client created with this method can only return primitive types without a model.
        """
        routes: dict[str, ClientEndpoint] = {}
        default_headers = {"User-Agent": f"BentoML HTTP Client/{__version__}"}
        if token is None:
            token = os.getenv("BENTO_CLOUD_API_KEY")
        if token:
            default_headers["Authorization"] = f"Bearer {token}"

        self._readyz_endpoint = "/readyz"

        if service is not None:
            for name, method in service.apis.items():
                routes[name] = ClientEndpoint(
                    name=name,
                    route=method.route,
                    input=method.input_spec.model_json_schema(),
                    output=method.output_spec.model_json_schema(),
                    doc=method.doc,
                    input_spec=method.input_spec,
                    output_spec=method.output_spec,
                    stream_output=method.is_stream,
                    is_task=method.is_task,
                )

            from bentoml._internal.context import server_context

            default_headers.update(
                {
                    "Bento-Name": server_context.bento_name,
                    "Bento-Version": server_context.bento_version,
                    "Runner-Name": service.name,
                    "Yatai-Bento-Deployment-Name": server_context.yatai_bento_deployment_name,
                    "Yatai-Bento-Deployment-Namespace": server_context.yatai_bento_deployment_namespace,
                }
            )
            self._readyz_endpoint = service.config.get("endpoints", {}).get(
                "readyz", self._readyz_endpoint
            )
        self.__attrs_init__(  # type: ignore
            url=url,
            endpoints=routes,
            media_type=media_type,
            default_headers=default_headers,
            timeout=timeout,
            app=app,
            server_ready_timeout=server_ready_timeout,
            service=service,
        )
        self._session_manager = SessionManager(
            url=url,
            timeout=timeout,
            headers=default_headers,
            app=app,
            max_age=max_age,
            max_requests=max_requests,
        )

    @_temp_dir.default  # type: ignore
    def default_temp_dir(self) -> tempfile.TemporaryDirectory[str]:
        return tempfile.TemporaryDirectory(prefix="bentoml-client-")

    @cached_property
    def serde(self) -> Serde:
        from ..serde import ALL_SERDE

        return ALL_SERDE[self.media_type]()

    async def close(self) -> None:
        await self._session_manager.close()
        self._file_manager.close()
        self._temp_dir.cleanup()

    def call(self, __name: str, /, *args: t.Any, **kwargs: t.Any) -> t.Any:
        try:
            endpoint = self.endpoints[__name]
        except KeyError:
            raise NotFound(f"Endpoint {__name} not found") from None
        if endpoint.stream_output:
            return self._get_stream(endpoint, args, kwargs)
        else:
            return self._call(endpoint, args, kwargs)

    async def _setup(self) -> None:
        if self._setup_done:
            return

        if self.app is None and (
            self.server_ready_timeout is None or self.server_ready_timeout > 0
        ):
            await self.wait_until_server_ready(self.server_ready_timeout)
        if self.service is None:
            schema_url = urljoin(self.url, "/schema.json")
            client = await self._session_manager.get_session()
            async with client.get("/schema.json") as resp:
                if not resp.ok:
                    raise BentoMLException(f"Failed to fetch schema from {schema_url}")
                for route in (await resp.json())["routes"]:
                    self.endpoints[route["name"]] = ClientEndpoint(
                        name=route["name"],
                        route=route["route"],
                        input=route["input"],
                        output=route["output"],
                        doc=route.get("doc"),
                        stream_output=route["output"].get("is_stream", False),
                        is_task=route.get("is_task", False),
                    )
        self._setup_endpoints()

    async def wait_until_server_ready(self, timeout: float | None = None) -> None:
        if timeout is None:
            timeout = self.timeout
        end = time.monotonic() + timeout
        while (current := time.monotonic()) < end:
            client = await self._session_manager.get_session()
            try:
                async with client.get(
                    self._readyz_endpoint, timeout=aiohttp.ClientTimeout(end - current)
                ) as resp:
                    if resp.status == 200:
                        return
            except (aiohttp.ClientError, aiohttp.ServerTimeoutError):
                pass
        raise ServiceUnavailable(f"Server is not ready after {timeout} seconds")

    async def is_ready(self, timeout: int | None = None) -> bool:
        client = await self._session_manager.get_session()
        try:
            resp = await client.get(
                self._readyz_endpoint,
                timeout=aiohttp.ClientTimeout(timeout) if timeout else None,
            )
            return resp.status == 200
        except aiohttp.ServerTimeoutError:
            logger.warning("Timed out waiting for runner to be ready")
            return False

    async def __aenter__(self) -> t.Self:
        await self._setup()
        return self

    async def __aexit__(self, *args: t.Any) -> None:
        return await self.close()

    async def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> AsyncTask:
        client = await self._session_manager.get_session()
        try:
            headers = CIMultiDict[str]()
            body = self._build_payload(__endpoint, args, kwargs, headers)
            url = f"{__endpoint.route}/submit"
            async with client.post(url, data=body, headers=headers) as resp:
                if not resp.ok:
                    raise BentoMLException(
                        f"Error making request: {resp.status}: {await resp.text()}",
                        error_code=HTTPStatus(resp.status),
                    )
            data = await resp.json()
            return AsyncTask(data["task_id"], __endpoint, self)
        finally:
            self._file_manager.close()

    async def _get_stream(
        self, endpoint: ClientEndpoint, args: t.Any, kwargs: t.Any
    ) -> t.AsyncGenerator[t.Any, None]:
        resp = await self._call(endpoint, args, kwargs)
        assert inspect.isasyncgen(resp)
        async for data in resp:
            yield data

    async def _call(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        *,
        headers: t.Mapping[str, str] | None = None,
    ) -> t.Any:
        client = await self._session_manager.get_session()
        try:
            headers = CIMultiDict({"Content-Type": self.media_type, **(headers or {})})
            body = self._build_payload(endpoint, args, kwargs, headers)
            resp = await client.post(endpoint.route, data=body, headers=headers)
            if not resp.ok:
                raise await map_exception(resp)
            if endpoint.stream_output:
                return self._parse_stream_response(endpoint, resp)
            elif endpoint.output.get("type") == "file":
                # file responses are always raw binaries whatever the serde is
                return await self._parse_file_response(endpoint, resp)
            else:
                return await self._parse_response(endpoint, resp)
        finally:
            self._file_manager.close()

    # ==========================
    # Request builders
    # ==========================

    def _build_payload(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        headers: dict[str, str],
    ) -> t.Any:
        from opentelemetry import propagate

        from _bentoml_sdk.io_models import IORootModel

        propagate.inject(headers)
        if endpoint.input_spec is not None:
            model = endpoint.input_spec.from_inputs(*args, **kwargs)
            if (
                not isinstance(model, IORootModel)
                and model.multipart_fields
                and self.media_type == "application/json"
            ):
                return self._build_multipart(endpoint, model, headers)
            elif isinstance(rendered := model.model_dump(), (str, bytes)):
                headers.update({"content-type": model.mime_type()})
                return rendered
            else:
                payload = self.serde.serialize_model(model)
                headers.update(payload.headers)
                return payload.aiter_bytes()

        assert self.media_type == "application/json", (
            "Non-JSON request is not supported"
        )
        if endpoint.input.get("root_input", False):
            if len(args) > 1 or kwargs:
                raise TypeError("Expected one positional argument for root input")
            if not args:
                return None
            value = args[0]
            passthrough = False
            content = None
            if "properties" in endpoint.input:
                kwargs = value
                args = ()
                passthrough = True
            elif endpoint.input.get("type") == "file":
                file = self._file_manager.get_file(value)
                if isinstance(file, str):
                    content = file
                else:
                    file_io, content_type = file[1:]
                    content = iter(lambda: file_io.read(4096), b"")
                    if content_type:
                        headers.update({"content-type": content_type})
            elif isinstance(value, (str, bytes)):
                content = value.encode("utf-8") if isinstance(value, str) else value
                headers.update({"content-type": "text/plain"})
            else:
                payload = self.serde.serialize(value, endpoint.input)
                headers.update(payload.headers)
                content = payload.aiter_bytes()
            if not passthrough:
                return content

        for name, value in zip(endpoint.input["properties"], args):
            if name in kwargs:
                raise TypeError(f"Duplicate argument {name}")
            kwargs[name] = value

        params = set(endpoint.input["properties"].keys())
        non_exist_args = set(kwargs.keys()) - set(params)
        if non_exist_args:
            raise TypeError(
                f"Arguments not found in endpoint {endpoint.name}: {non_exist_args}"
            )
        required = set(endpoint.input.get("required", []))
        missing_args = set(required) - set(kwargs.keys())
        if missing_args:
            raise TypeError(
                f"Missing required arguments in endpoint {endpoint.name}: {missing_args}"
            )
        has_file = any(
            schema.get("type") == "file"
            or schema.get("type") == "array"
            and schema["items"].get("type") == "file"
            for schema in endpoint.input["properties"].values()
        )
        if has_file:
            return self._build_multipart(endpoint, kwargs, headers)
        payload = self.serde.serialize(kwargs, endpoint.input)
        headers.update(payload.headers)
        return payload.aiter_bytes()

    def _build_multipart(
        self,
        endpoint: ClientEndpoint,
        model: IODescriptor | dict[str, t.Any],
        headers: dict[str, str],
    ) -> aiohttp.FormData:
        def is_file_field(k: str) -> bool:
            if isinstance(model, IODescriptor):
                return k in model.multipart_fields
            if (f := endpoint.input["properties"].get(k, {})).get("type") == "file":
                return True
            if f.get("type") == "array" and f["items"].get("type") == "file":
                return True
            return False

        if isinstance(model, dict):
            fields = model
        else:
            fields = {t.cast(str, k): getattr(model, k) for k in model.model_fields}
        data: dict[str, t.Any] = {}
        files: list[tuple[str, tuple[str, t.IO[bytes], str | None]]] = []

        for name, value in fields.items():
            if not is_file_field(name):
                data[name] = json.dumps(value)
                continue
            if not isinstance(value, (list, tuple)):
                value = [value]

            for v in value:
                file = self._file_manager.get_file(v)
                if isinstance(file, str):
                    data[name] = file
                else:
                    files.append((name, file))
        headers.pop("content-type", None)
        payload = aiohttp.FormData()
        for key, val in data.items():
            payload.add_field(key, val)
        for key, (filename, fileobj, content_type) in files:
            payload.add_field(
                key,
                fileobj,
                filename=filename,
                content_type=content_type,
            )
        return payload

    # ==========================
    # Response parsers
    # ==========================

    def _deserialize_output(self, payload: Payload, endpoint: ClientEndpoint) -> t.Any:
        from _bentoml_sdk.io_models import IORootModel

        data = payload.iter_bytes()
        if (endpoint.output.get("type")) == "string":
            content = bytes(next(data))
            if endpoint.output.get("format") == "binary":
                return content
            return content.decode("utf-8")
        elif endpoint.output_spec is not None:
            model = self.serde.deserialize_model(payload, endpoint.output_spec)
            if isinstance(model, IORootModel):
                return model.root  # type: ignore
            return model
        else:
            return self.serde.deserialize(payload, endpoint.output)

    async def _parse_response(
        self, endpoint: ClientEndpoint, resp: aiohttp.ClientResponse
    ) -> t.Any:
        data = await resp.read()
        return self._deserialize_output(Payload((data,), resp.headers), endpoint)

    async def _parse_stream_response(
        self, endpoint: ClientEndpoint, resp: aiohttp.ClientResponse
    ) -> t.AsyncGenerator[t.Any, None]:
        async for data in resp.content.iter_chunked(1024 * 8):
            yield self._deserialize_output(Payload((data,), resp.headers), endpoint)

    async def _parse_file_response(
        self, endpoint: ClientEndpoint, resp: aiohttp.ClientResponse
    ) -> pathlib.Path | Image.Image:
        from PIL import Image
        from python_multipart.multipart import parse_options_header

        content_disposition = resp.headers.get("content-disposition")
        content_type = resp.headers.get("content-type", "")
        filename: str | None = None
        if endpoint.output.get("pil"):
            image_formats = (
                [content_type[6:]] if content_type.startswith("image/") else None
            )
            return Image.open(io.BytesIO(await resp.read()), formats=image_formats)
        if content_disposition:
            _, options = parse_options_header(content_disposition)
            if b"filename" in options:
                filename = str(options[b"filename"], "utf8", errors="ignore")

        with tempfile.NamedTemporaryFile(
            "wb", suffix=filename, dir=self._temp_dir.name, delete=False
        ) as f:
            f.write(await resp.read())
        return pathlib.Path(f.name)

    async def _get_task_status(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> ResultStatus:
        client = await self._session_manager.get_session()
        async with client.get(
            f"{__endpoint.route}/status", params={"task_id": task_id}
        ) as resp:
            if not resp.ok:
                raise await map_exception(resp)
            data = await resp.json()
            return ResultStatus(data["status"])

    async def _cancel_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> None:
        client = await self._session_manager.get_session()
        async with client.put(
            f"{__endpoint.route}/cancel", params={"task_id": task_id}
        ) as resp:
            if not resp.ok:
                raise await map_exception(resp)

    async def _retry_task(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> AsyncTask:
        client = await self._session_manager.get_session()
        async with client.post(
            f"{__endpoint.route}/retry", params={"task_id": task_id}
        ) as resp:
            if not resp.ok:
                raise await map_exception(resp)
            data = await resp.json()
            return AsyncTask(data["task_id"], __endpoint, self)

    async def _get_task_result(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> t.Any:
        client = await self._session_manager.get_session()
        async with client.get(
            f"{__endpoint.route}/get", params={"task_id": task_id}
        ) as resp:
            if not resp.ok:
                raise await map_exception(resp)
            if (
                __endpoint.output.get("type") == "file"
                and self.media_type == "application/json"
            ):
                return await self._parse_file_response(__endpoint, resp)
            else:
                return await self._parse_response(__endpoint, resp)


class SyncClient(AbstractClient):
    def __init__(self, async_client: AsyncClient) -> None:
        self._async_client = async_client

    @staticmethod
    def run_async(
        callable: t.Callable[P, t.Awaitable[T]], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        return anyio.from_thread.run(functools.partial(callable, *args, **kwargs))

    def call(self, __name: str, /, *args: t.Any, **kwargs: t.Any) -> t.Any:
        coro = self._async_client.call(__name, *args, **kwargs)
        if inspect.isasyncgen(coro):

            def generator() -> t.Generator[t.Any, None, None]:
                while True:
                    try:
                        yield anyio.from_thread.run(coro.__anext__)
                    except StopAsyncIteration:
                        break

            return generator()

        return self.run_async(lambda: coro)

    def _get_task_status(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        return self.run_async(self._async_client._get_task_status, __endpoint, task_id)

    def _cancel_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> None:
        return self.run_async(self._async_client._cancel_task, __endpoint, task_id)

    def _retry_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        return self.run_async(self._async_client._retry_task, __endpoint, task_id)

    def _get_task_result(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        return self.run_async(self._async_client._get_task_result, __endpoint, task_id)

    def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> Task:
        async_task = self.run_async(
            self._async_client._submit, __endpoint, *args, **kwargs
        )
        return Task(async_task.id, __endpoint, self)


class RemoteProxy(AbstractClient, t.Generic[T]):
    """A remote proxy of the passed in service that has the same interfaces"""

    def __init__(
        self,
        url: str,
        *,
        service: Service[T] | None = None,
        media_type: str = "application/vnd.bentoml+pickle",
        app: ASGIApp | None = None,
    ) -> None:
        from bentoml.container import BentoMLContainer

        if service is not None:
            svc_config: dict[str, ServiceConfig] = (
                BentoMLContainer.config.services.get()
            )
            timeout = (
                svc_config.get(service.name, {}).get("traffic", {}).get("timeout") or 60
            ) * 1.01  # get the service timeout add 1% margin for the client
            runner_conn_config = svc_config.get(service.name, {}).get(
                "runner_connection", {}
            )
            max_age = runner_conn_config.get("max_age", 300.0)
            max_requests = runner_conn_config.get("max_requests", 100)
        else:
            timeout = 60
            max_age = 300.0
            max_requests = 100
        self._async = AsyncClient(
            url,
            media_type=media_type,
            service=service,
            timeout=timeout,
            server_ready_timeout=0,
            app=app,
            max_age=max_age,
            max_requests=max_requests,
        )
        self._sync = SyncClient(self._async)
        if service is not None:
            self._inner = service.inner
        else:
            self._inner = None

    async def __aenter__(self) -> t.Self:
        await self._async.__aenter__()
        self.endpoints = self._sync.endpoints = self._async.endpoints
        self._sync._setup_endpoints()
        self._setup_endpoints()
        return self

    @property
    def to_async(self) -> AsyncClient:
        return self._async

    @property
    def to_sync(self) -> SyncClient:
        return self._sync

    @property
    def client_url(self) -> str:
        return str(self._async.client._base_url)  # type: ignore[attr-defined]

    async def is_ready(self, timeout: int | None = None) -> bool:
        return await self._async.is_ready(timeout=timeout)

    async def close(self) -> None:
        await self._async.close()

    def as_service(self) -> T:
        return t.cast(T, self)

    def call(self, __name: str, /, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if self._inner is None:
            raise BentoMLException(
                "The proxy is not callable when the service is not provided. Please use `.to_async` or `.to_sync` property."
            )
        original_func = getattr(self._inner, __name)
        if not hasattr(original_func, "func"):
            raise BentoMLException(f"calling non-api method {__name} is not allowed")
        original_func = original_func.func
        while isinstance(original_func, functools.partial):
            original_func = original_func.func
        is_async_func = (
            asyncio.iscoroutinefunction(original_func)
            or (
                callable(original_func)
                and asyncio.iscoroutinefunction(original_func.__call__)  # type: ignore
            )
            or inspect.isasyncgenfunction(original_func)
        )
        if is_async_func:
            return self._async.call(__name, *args, **kwargs)
        else:
            return self._sync.call(__name, *args, **kwargs)

    def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> t.Any:
        original_func = getattr(self._inner, __endpoint.name)
        if not hasattr(original_func, "func"):
            raise BentoMLException(
                f"calling non-api method {__endpoint.name} is not allowed"
            )
        original_func = original_func.func
        while isinstance(original_func, functools.partial):
            original_func = original_func.func
        is_async_func = (
            asyncio.iscoroutinefunction(original_func)
            or (
                callable(original_func)
                and asyncio.iscoroutinefunction(original_func.__call__)  # type: ignore
            )
            or inspect.isasyncgenfunction(original_func)
        )
        if is_async_func:
            return self._async._submit(__endpoint, *args, **kwargs)
        else:
            return self._sync._submit(__endpoint, *args, **kwargs)



"""
User facing python APIs for API token management
"""

from __future__ import annotations

import typing as t
from datetime import datetime

from simple_di import Provide
from simple_di import inject

from ._internal.cloud.api_token import ApiToken
from ._internal.configuration.containers import BentoMLContainer

if t.TYPE_CHECKING:
    from ._internal.cloud import BentoCloudClient


@inject
def list(
    search: str | None = None,
    _cloud_client: "BentoCloudClient" = Provide[BentoMLContainer.bentocloud_client],
) -> t.List[ApiToken]:
    """List all API tokens.

    Args:
        search: Optional search string to filter tokens by name

    Returns:
        List of ApiToken objects

    Example:
        >>> import bentoml
        >>> tokens = bentoml.api_token.list(search="my-token")
        >>> for token in tokens:
        ...     print(f"{token.name}: {token.uid}")
    """
    return _cloud_client.api_token.list(search=search)


@inject
def create(
    name: str,
    description: str | None = None,
    scopes: t.List[str] | None = None,
    expired_at: datetime | None = None,
    _cloud_client: "BentoCloudClient" = Provide[BentoMLContainer.bentocloud_client],
) -> ApiToken:
    """Create a new API token.

    Args:
        name: Name of the token
        description: Optional description
        scopes: List of scopes. Available scopes:
            - api: General API access
            - read_organization: Read organization data
            - write_organization: Write organization data
            - read_cluster: Read cluster data
            - write_cluster: Write cluster data
        expired_at: Optional expiration datetime

    Returns:
        ApiToken object (includes token value - save it, it won't be shown again!)

    Example:
        >>> import bentoml
        >>> from datetime import datetime, timedelta
        >>> token = bentoml.api_token.create(
        ...     name="ci-token",
        ...     description="CI/CD pipeline token",
        ...     scopes=["api", "read_cluster"],
        ...     expired_at=datetime.now() + timedelta(days=30)
        ... )
        >>> print(f"Token: {token.token}")  # Save this!
    """
    return _cloud_client.api_token.create(
        name=name,
        description=description,
        scopes=scopes,
        expired_at=expired_at,
    )


@inject
def get(
    token_uid: str,
    _cloud_client: "BentoCloudClient" = Provide[BentoMLContainer.bentocloud_client],
) -> ApiToken | None:
    """Get an API token by UID.

    Args:
        token_uid: The UID of the token

    Returns:
        ApiToken object or None if not found

    Example:
        >>> import bentoml
        >>> token = bentoml.api_token.get("token_abc123")
        >>> if token:
        ...     print(f"Scopes: {token.scopes}")
    """
    return _cloud_client.api_token.get(token_uid=token_uid)


@inject
def delete(
    token_uid: str,
    _cloud_client: "BentoCloudClient" = Provide[BentoMLContainer.bentocloud_client],
) -> None:
    """Delete an API token.

    Args:
        token_uid: The UID of the token to delete

    Example:
        >>> import bentoml
        >>> bentoml.api_token.delete("token_abc123")
    """
    _cloud_client.api_token.delete(token_uid=token_uid)


__all__ = ["create", "get", "list", "delete"]
